"""
Module for detecting various quality control aspects from source data.
"""

import operator
import pathlib
from functools import reduce
from typing import Dict, List, Optional, Union

import pandas as pd
import yaml
from cytodataframe.frame import CytoDataFrame
from scipy.stats import zscore as scipy_zscore

DEFAULT_QC_THRESHOLD_FILE = (
    f"{pathlib.Path(__file__).parent!s}/data/qc_nuclei_thresholds_default.yml"
)


def identify_outliers(
    df: Union[CytoDataFrame, pd.DataFrame, str],
    feature_thresholds: Union[Dict[str, float], str],
    feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
    include_threshold_scores: bool = False,
    export_path: Optional[str] = None,
) -> Union[pd.Series, CytoDataFrame]:
    """
    This function uses z-scoring to format the data for detecting outlier
    nuclei or cells using specific CellProfiler features. Thresholds are
    the number of standard deviations away from the mean, either above
    (positive) or below (negative). We recommend making sure to not use a
    threshold of 0 as that would represent the whole dataset.

    Args:
        df: Union[CytoDataFrame, pd.DataFrame, str]
            DataFrame or file string-based filepath of a
            Parquet, CSV, or TSV file with CytoTable output or similar data.
        feature_thresholds: Dict[str, float]
            One of two options:
            A dictionary with the feature name(s) as the key(s) and their assigned
            threshold for identifying outliers. Positive int for the threshold
            will detect outliers "above" than the mean, negative int will detect
            outliers "below" the mean.
            Or a string which is a named key reference found within
            the feature_thresholds_file yaml file.
        feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
            An optional feature thresholds file where thresholds may be
            defined within a file.
        include_threshold_scores: bool
            Whether to include the threshold scores in addition to whether
            the threshold set passes per row.
        export_path: Optional[str] = None
            An optional path to export the data using CytoDataFrame export
            capabilities. If None no export is performed.
            Note: compatible exports are CSV's, TSV's, and parquet.

    Returns:
        Union[pd.Series, CytoDataFrame]:
            Outlier series with booleans based on whether outliers were detected
            or not for use within other functions.
    """

    # Ensure the input is a CytoDataFrame, converting if necessary
    df = CytoDataFrame(data=df)

    # reference the df for a new outlier_df
    outlier_df = df

    # Define the naming scheme for z-score columns based on thresholds
    thresholds_name = (
        f"cqc.{feature_thresholds}"
        if isinstance(feature_thresholds, str)
        else "cqc.custom"
    )

    # If feature_thresholds is a string, load the thresholds from the specified file
    if isinstance(feature_thresholds, str):
        feature_thresholds = read_thresholds_set_from_file(
            feature_thresholds=feature_thresholds,
            feature_thresholds_file=feature_thresholds_file,
        )

    # Dictionary to store mappings of features to their z-score column names
    zscore_columns = {}
    for feature in feature_thresholds:
        # Ensure the feature exists in the DataFrame
        if feature not in df.columns:
            raise ValueError(f"Feature '{feature}' does not exist in the DataFrame.")

        # Construct the z-score column name
        zscore_col = f"{thresholds_name}.Z_Score.{feature}"

        # Calculate and store z-scores only if not already present
        if zscore_col not in outlier_df:
            outlier_df[zscore_col] = scipy_zscore(df[feature])

        # Add the column name to the zscore_columns dictionary
        zscore_columns[feature] = zscore_col

    # Helper function to create outlier detection conditions
    def create_condition(feature: str, threshold: float) -> pd.Series:
        # Positive threshold checks for outliers above the mean
        if threshold > 0:
            return outlier_df[zscore_columns[feature]] > threshold
        # Negative threshold checks for outliers below the mean
        return outlier_df[zscore_columns[feature]] < threshold

    # Generate outlier detection conditions for all features
    conditions = [
        create_condition(feature, threshold)
        for feature, threshold in feature_thresholds.items()
    ]

    # Construct the result based on whether threshold scores should be included
    if include_threshold_scores:
        # Extract z-score columns for each feature
        zscore_df = outlier_df[list(zscore_columns.values())]

        # Combine conditions into a single Series indicating outlier status
        is_outlier_series = reduce(operator.and_, conditions).rename(
            f"{thresholds_name}.is_outlier"
        )

        # Combine z-scores and outlier status into a single DataFrame
        result = CytoDataFrame(
            data=pd.concat([zscore_df, is_outlier_series], axis=1),
            data_context_dir=df._custom_attrs["data_context_dir"],
            data_mask_context_dir=df._custom_attrs["data_mask_context_dir"],
        )
    else:
        # Combine conditions into a single Series of boolean values
        result = reduce(operator.and_, conditions)

    # Export the result if an export path is specified
    if export_path is not None:
        export_df = CytoDataFrame(result) if isinstance(result, pd.Series) else result
        export_df.export(file_path=export_path)

    # Return the resulting Series or DataFrame
    return result


def find_outliers(
    df: Union[CytoDataFrame, pd.DataFrame, str],
    metadata_columns: List[str],
    feature_thresholds: Union[Dict[str, float], str],
    feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
    export_path: Optional[str] = None,
) -> pd.DataFrame:
    """
    This function uses identify_outliers to return a dataframe
    with only the outliers and provided metadata columns.

    Args:
        df: Union[CytoDataFrame, pd.DataFrame, str]
            DataFrame or file string-based filepath of a
            Parquet, CSV, or TSV file with CytoTable output or similar data.
        metadata_columns: List[str]
            List of metadata columns that should be outputted with the outlier data.
        feature_thresholds: Dict[str, float]
            One of two options:
            A dictionary with the feature name(s) as the key(s) and their assigned
            threshold for identifying outliers. Positive int for the threshold
            will detect outliers "above" than the mean, negative int will detect
            outliers "below" the mean.
            Or a string which is a named key reference found within
            the feature_thresholds_file yaml file.
        feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
            An optional feature thresholds file where thresholds may be
            defined within a file.
        export_path: Optional[str] = None
            An optional path to export the data using CytoDataFrame export
            capabilities. If None no export is performed.
            Note: compatible exports are CSV's, TSV's, and parquet.

    Returns:
        pd.DataFrame:
            Outlier data frame for the given conditions.
    """

    # Resolve feature_thresholds if provided as a string
    if isinstance(feature_thresholds, str):
        feature_thresholds = read_thresholds_set_from_file(
            feature_thresholds=feature_thresholds,
            feature_thresholds_file=feature_thresholds_file,
        )

    # Determine the columns required for processing
    required_columns = list(feature_thresholds.keys()) + metadata_columns

    # Interpret the df as CytoDataFrame
    df = CytoDataFrame(data=df)[required_columns]

    # if we have nan's in our columns, emit a warning and drop them
    if any(df[list(feature_thresholds.keys())].isna().any()):
        print(
            "Warning: NaN values found in the DataFrame. "
            "These will be dropped before processing."
        )
        df = df.dropna(subset=list(feature_thresholds.keys()))

    # Filter DataFrame for outliers using identify_outliers
    outliers_mask = identify_outliers(
        # Select only the required columns from the DataFrame
        df=df,
        feature_thresholds=feature_thresholds,
        feature_thresholds_file=feature_thresholds_file,
    )
    outliers_df = df[outliers_mask]

    # Print outlier count and range for each feature
    print(
        "Number of outliers:",
        outliers_df.shape[0],
        f"({'{:.2f}'.format((outliers_df.shape[0] / df.shape[0]) * 100)}%)",
    )
    print("Outliers Range:")
    for feature in feature_thresholds:
        print(f"{feature} Min:", outliers_df[feature].min())
        print(f"{feature} Max:", outliers_df[feature].max())

    # Include metadata columns in the output DataFrame
    result = outliers_df[required_columns]

    # Export the file if specified
    if export_path is not None:
        result.export(file_path=export_path)

    # Return the resulting DataFrame
    return result


def label_outliers(
    df: Union[CytoDataFrame, pd.DataFrame, str],
    feature_thresholds: Optional[Union[Dict[str, float], str]] = None,
    feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
    include_threshold_scores: bool = False,
    export_path: Optional[str] = None,
) -> CytoDataFrame:
    """
    Use identify_outliers to label the original dataset for
    where a cell passed or failed the quality control condition(s).

        Args:
            df: Union[CytoDataFrame, pd.DataFrame, str]
                DataFrame or file string-based filepath of a
                Parquet, CSV, or TSV file with CytoTable output or similar data.
            feature_thresholds: Dict[str, float]
                One of two options:
                A dictionary with the feature name(s) as the key(s) and their assigned
                threshold for identifying outliers. Positive int for the threshold
                will detect outliers "above" than the mean, negative int will detect
                outliers "below" the mean.
                Or a string which is a named key reference found within
                the feature_thresholds_file yaml file.
            feature_thresholds_file: Optional[str] = DEFAULT_QC_THRESHOLD_FILE,
                An optional feature thresholds file where thresholds may be
                defined within a file.
            include_threshold_scores: bool = False
                Whether to include the scores in addition to whether an outlier
                was detected or not.
            export_path: Optional[str] = None
                An optional path to export the data using CytoDataFrame export
                capabilities. If None no export is performed.
                Note: compatible exports are CSV's, TSV's, and parquet.

        Returns:
            CytoDataFrame:
                Full dataframe with optional scores and outlier boolean column.
    """

    # interpret the df as CytoDataFrame
    if not isinstance(df, CytoDataFrame):
        df = CytoDataFrame(data=df)

    # store the custom attributes
    custom_attrs = dict(df._custom_attrs)

    # for single outlier processing
    if isinstance(feature_thresholds, (str, dict)):
        # return the outlier dataframe for one threshold rule
        identified_outliers = identify_outliers(
            df=df,
            feature_thresholds=feature_thresholds,
            feature_thresholds_file=feature_thresholds_file,
            include_threshold_scores=include_threshold_scores,
        )

        result = CytoDataFrame(
            data=pd.concat(
                [
                    df,
                    (
                        identified_outliers
                        if isinstance(identified_outliers, pd.DataFrame)
                        else CytoDataFrame(
                            {
                                (
                                    f"cqc.{feature_thresholds}.is_outlier"
                                    if isinstance(feature_thresholds, str)
                                    else "cqc.custom.is_outlier"
                                ): identified_outliers
                            }
                        )
                    ),
                ],
                axis=1,
            ),
            # reuse the custom attributes
            **custom_attrs,
        )

    # for multiple outlier processing
    elif feature_thresholds is None:
        # return the outlier dataframe for all threshold rules
        labeled_df = pd.concat(
            [df]
            + [
                # identify outliers for each threshold rule
                identify_outliers(
                    df=df,
                    feature_thresholds=thresholds,
                    feature_thresholds_file=feature_thresholds_file,
                    include_threshold_scores=include_threshold_scores,
                )
                # loop through each threshold rule
                for thresholds in read_thresholds_set_from_file(
                    feature_thresholds_file=feature_thresholds_file,
                )
            ],
            axis=1,
        )

        # return a dataframe with deduplicated columns by name
        result = CytoDataFrame(
            labeled_df.loc[:, ~labeled_df.columns.duplicated()],
            # reuse the custom attributes
            **custom_attrs,
        )

    # export the file if specified
    if export_path is not None:
        result.export(file_path=export_path)

    return result


def read_thresholds_set_from_file(
    feature_thresholds_file: str, feature_thresholds: Optional[str] = None
) -> Union[Dict[str, int], Dict[str, Dict[str, int]]]:
    """
    Reads a set of feature thresholds from a specified file.

    This function takes the path to a feature thresholds file and a
    specific feature threshold string, reads the file, and returns
    the thresholds set from the file.

    Args:
        feature_thresholds_file (str):
            The path to the file containing feature thresholds.
        feature_thresholds (Optional str, default None):
            A string specifying the feature thresholds.
            If we have None, return all thresholds.

    Returns:
        dict: A dictionary containing the processed feature thresholds.

    Raises:
        LookupError: If the file does not contain the specified feature_thresholds key.
    """

    # open the yaml file
    with open(feature_thresholds_file, "r") as file:
        thresholds = yaml.safe_load(file)

    # if no feature thresholds name is specified, return all thresholds
    if feature_thresholds is None:
        return thresholds["thresholds"]

    if feature_thresholds not in thresholds["thresholds"]:
        raise LookupError(
            (
                f"Unable to find threshold set by name {feature_thresholds}"
                f" within {feature_thresholds_file}"
            )
        )

    return thresholds["thresholds"][feature_thresholds]



"""
Setup coSMicQC CLI through python-fire
"""

import inspect
import sys
import types
from typing import Optional

import fire
from fire import helptext, inspectutils, value_types
from fire.core import Display, FireError, _DictAsString, _OneLineResult
from fire.trace import FireTrace

from . import analyze


# used to avoid bugs with python-fire and pandas string-based repr compatibility.
# referenced from https://github.com/google/python-fire/pull/446
# to be removed after python-fire merges changes (uncertain of timeline)
def HasCustomRepr(component: object) -> bool:
    """Reproduces above HasCustomStr function to determine if component has a
    custom __repr__ method.

    ...

    Args:
      component: The object to check for a custom __repr__ method.
    Returns:
      Whether `component` has a custom __repr__ method.
    """
    if hasattr(component, "__repr__"):
        class_attrs = inspectutils.GetClassAttrsDict(type(component)) or {}
        repr_attr = class_attrs.get("__repr__")
        if repr_attr and repr_attr.defining_class is not object:
            return True
    return False


# used to avoid bugs with python-fire and pandas string-based repr compatibility.
# referenced with modifications from https://github.com/google/python-fire/pull/446
# to be removed after python-fire merges changes (uncertain of timeline)
# ignore rule below added to help avoid triggering ruff linting checks on temporary fix.
def _PrintResult(  # noqa: C901
    component_trace: FireTrace, verbose: bool = False, serialize: Optional[bool] = None
) -> None:
    """Prints the result of the Fire call to stdout in a human readable way."""
    result = component_trace.GetResult()
    # Allow users to modify the return value of the component and provide
    # custom formatting.
    if serialize:
        if not callable(serialize):
            raise FireError(
                "The argument `serialize` must be empty or callable:", serialize
            )
        result = serialize(result)
    if value_types.HasCustomStr(result):
        # If the object has a custom __str__ method, rather than one inherited from
        # object, then we use that to serialize the object.
        print(str(result))
        return

    elif HasCustomRepr(result):
        # Same as above, but for __repr__.
        # For pandas.DataFrame, __str__ is inherited from object, but __repr__ has
        # a custom implementation (see pandas.core.frame.DataFrame.__repr__)
        print(str(result))
        return

    if isinstance(result, (list, set, frozenset, types.GeneratorType)):
        for i in result:
            print(_OneLineResult(i))
    elif inspect.isgeneratorfunction(result):
        raise NotImplementedError
    elif isinstance(result, dict) and value_types.IsSimpleGroup(result):
        print(_DictAsString(result, verbose))
    elif isinstance(result, tuple):
        print(_OneLineResult(result))
    elif isinstance(result, value_types.VALUE_TYPES):
        if result is not None:
            print(result)
    else:
        help_text = helptext.HelpText(result, trace=component_trace, verbose=verbose)
        output = [help_text]
        # used for displaying output through python-fire
        Display(output, out=sys.stdout)


# replace the _PrintResult function with a fix for pandas dataframes
fire.core._PrintResult = _PrintResult


def cli_analyze() -> None:
    """
    Run the analyze module functions through python-fire CLI

    This function serves as the CLI entry point for functions
    within the analyze module.
    """

    fire.Fire(analyze)


if __name__ == "__main__":
    """
    Setup the CLI with python-fire for the coSMicQC package.

    This enables running the functions identify_outliers, find_outliers,
    and label_outliers from the command line interface through analyze
    """

    cli_analyze()



"""
Module for detecting contamination (e.g., mycoplasma) from the morphology profiles.
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from tabulate import tabulate

from cosmicqc import find_outliers


def bool_to_emoji(val: bool) -> str:
    """Convert boolean string to emoji for printing.

    Args:
        val (bool): boolean value to convert to emoji

    Returns:
        str: emoji representing true or false
    """
    return "✅" if val else "❌"


class ContaminationDetector:
    """
    This class implements a contamination detection process for
    high-content morphology data.
    Contamination can be classified as many things, including
    mycoplasma, which can impact the morphology of cells.
    There are methods in the wet lab to detect mycoplasma,
    but it is not always full-proof.
    This class includes methods to detect contamination of
    any form based on nucleus morphology features.
    There are three steps to the process:

    1. Check for skewness in the cytoplasm texture and variability
    in the FormFactor features. If either is detected, the process
    moves on to the next step.
    2. Determine if the whole plate or partial plate is impacted by
    checking the mean values of the features.
    The process will move on to the last step only if partial
    contamination is detected in the texture feature.
    3. If partial contamination is detected for texture, find outliers
    and plot the proportion of outliers per well.

    Attributes:
        dataframe : pd.DataFrame
            The input DataFrame containing the features to be analyzed.
            Should be the output from the CytoTable convert function.
        nucleus_channel_naming : str
            The naming convention for the nucleus channel in the DataFrame.
        cyto_feature : str
            The feature name for cytoplasm texture.
        formfactor_feature : str
            The feature name for FormFactor.
        lower_skew_threshold : float
            Lower threshold for Bowley's skewness.
        upper_skew_threshold : float
            Upper threshold for Bowley's skewness.
        variability_threshold : float
            Threshold for variability in the FormFactor feature.
        texture_mean_threshold : float
            Threshold for mean value of cytoplasm texture.
        formfactor_mean_threshold : float
            Threshold for mean value of FormFactor.
        outlier_std_threshold : float
            Standard deviation threshold for outlier detection.
        is_skewed : bool
            Indicates if the distribution is skewed.
        is_variable : bool
            Indicates if the distribution is variable.
        whole_plate_contamination_texture : bool
            Indicates if whole plate contamination is detected in texture feature.
        whole_plate_contamination_formfactor : bool
            Indicates if whole plate contamination is detected in FormFactor feature.
        partial_contamination_texture_detected : bool
            Indicates if partial contamination is detected in the texture feature.
    """

    def __init__(  # noqa: PLR0913
        self,
        dataframe: pd.DataFrame,
        nucleus_channel_naming: str = "DNA",
        lower_skew_threshold: float = -0.15,
        upper_skew_threshold: float = 0.09,
        variability_threshold: float = 0.15,
        texture_mean_threshold: float = -0.25,
        formfactor_mean_threshold: float = 0.78,
        outlier_std_threshold: float = 1.0,
    ) -> None:
        """
        Initializer for the ContaminationDetector class.

        Args:
            dataframe (pd.DataFrame): The input DataFrame containing the features
                to be analyzed. Should be the output from CytoTable convert.
            nucleus_channel_naming (str): Naming convention for the nucleus
                channel in the DataFrame. Defaults to "DNA".
            lower_skew_threshold (float): Lower threshold for Bowley's
                skewness to flag whole-plate contamination. Defaults to -0.15.
            upper_skew_threshold (float): Upper threshold for Bowley's
                skewness to flag partial-plate contamination. Defaults to 0.09.
            variability_threshold (float): Threshold for acceptable
                coefficient of variation in selected features. Defaults to 0.15.
            texture_mean_threshold (float): Mean texture value threshold
                for detecting abnormalities. Defaults to -0.25.
            formfactor_mean_threshold (float): Mean FormFactor threshold
                for detecting nuclear shape issues. Defaults to 0.78.
            outlier_std_threshold (float): Number of standard deviations from
                the mean used to flag outliers. Defaults to 1.0.
        """
        self.dataframe = dataframe
        self.nucleus_channel_naming = nucleus_channel_naming
        self.lower_skew_threshold = lower_skew_threshold
        self.upper_skew_threshold = upper_skew_threshold
        self.variability_threshold = variability_threshold
        self.texture_mean_threshold = texture_mean_threshold
        self.formfactor_mean_threshold = formfactor_mean_threshold
        self.outlier_std_threshold = outlier_std_threshold

        # set the features to be used for contamination detection
        self.cyto_feature = (
            f"Cytoplasm_Texture_InfoMeas1_{self.nucleus_channel_naming}_3_02_256"
        )
        self.formfactor_feature = "Nuclei_AreaShape_FormFactor"

    def _skewness_test_cytoplasm_texture(self) -> bool:
        """
        Bowley's skewness score is calculated for the cytoplasm texture
        around the nucleus feature.
        Thresholds are assigned to determine if there is abnormal texture around
        the nucleus, either based on a serious negative skew
        (indicating a whole plate issue) or a positive skew
        (indicating a partial plate issue).
        Thresholds are set based on findings from multiple experiments
        (NF1 & CFReT, see citation file).

        Reference: Bowley, A. L. (1901): Elements of Statistics, P.S. King and Son, Laondon.

        Returns:
            boolean: True if the distribution is skewed, False otherwise.
        """  # noqa: E501
        # calculate the quartiles
        q1 = np.percentile(self.dataframe[self.cyto_feature].dropna(), 25)
        q2 = np.percentile(self.dataframe[self.cyto_feature].dropna(), 50)
        q3 = np.percentile(self.dataframe[self.cyto_feature].dropna(), 75)

        # calculate Bowley's skewness (to detect partial contamination)
        bowley_skewness = (q3 + q1 - 2 * q2) / (q3 - q1)

        # check if the distribution is skewed
        # Bowley's skewness is negative for a whole plate issue
        # and positive for a partial plate issue
        return (
            bowley_skewness < self.lower_skew_threshold
            or bowley_skewness > self.upper_skew_threshold
        )

    def _variability_test_formfactor(self) -> bool:
        """
        Calculate Interquartile Range (IQR) score for the FormFactor feature
        to determine if there is abnormal variability in the shape of the nuclei.
        More variability in this feature indicates a higher proportion
        of non-circular nuclei.
        This can be an indicator of poor segmentations due to non-optimal
        segmentation parameters or contamination.
        Threshold is set based on findings from multiple experiments
        (NF1 & CFReT, see citation file).

        Returns:
            boolean: True if the distribution is variable, False otherwise.
        """
        # use IQR to measure variability of the feature
        vals = self.dataframe[self.formfactor_feature].dropna()
        iqr = np.percentile(vals, 75) - np.percentile(vals, 25)

        # determine if distribution is highly variable
        return iqr > self.variability_threshold

    def check_skew_and_variable(self) -> None:
        """
        Step 1: Check for skewness in the cytoplasm texture around the nucleus and
        variability in the FormFactor of the nucleus.
        """
        print("Running step 1...")

        # perform skewness and variability tests
        self.is_skewed = self._skewness_test_cytoplasm_texture()
        self.is_variable = self._variability_test_formfactor()

        # format results in a pretty table
        print("Summary:\n")
        print(
            tabulate(
                [
                    ["Texture skewed?", bool_to_emoji(self.is_skewed)],
                    ["Nucleus shape variable?", bool_to_emoji(self.is_variable)],
                ],
                headers=["Check", "Result"],
                tablefmt="simple",
            )
        )

        # set interpretation based on results
        if self.is_skewed and self.is_variable:
            interpretation = (
                "Contamination detected! 🚨\n"
                "Anomalous texture around nuclei detected and "
                "nuclei segmentation is impacted.\n"
                "Proceeding to step 2..."
            )
        elif self.is_skewed:
            interpretation = (
                "Contamination detected! 🚨\n"
                "Anomalous texture around nuclei detected but "
                "nuclei segmentation not clearly impacted.\n"
                "Proceeding to step 2..."
            )
        elif self.is_variable:
            interpretation = (
                "Potential contamination detected! 🛑\n"
                "This could indicate contamination, segmentation issues, "
                "or an interesting phenotype.\n"
                "Proceeding to step 2..."
            )
        else:
            interpretation = "No indication of contamination. Plate appears clean 🫧!"

        print(f"Interpretation:\n{interpretation}")

    def _calculate_texture_mean(self) -> bool:
        """
        Check the mean value of cytoplasm texture around a nucleus feature to
        determine if whole plate or partial plate contamination is present.
        Threshold is set based on findings from multiple experiments
        (NF1 & CFReT, see citation file).

        Normal raw values of this feature range below 0. In a clean plate,
        it is noted that the normal values range from -0.5 to -0.3.
        We set the threshold to -0.25 to detect whole plate contamination.

        Returns:
            boolean: True if whole plate contamination is detected, False otherwise
            indicates partial contamination.
        """
        # check if the mean value is above the threshold
        # True = whole plate, False = partial plate
        return self.dataframe[self.cyto_feature].mean() >= self.texture_mean_threshold

    def _calculate_formfactor_mean(self) -> bool:
        """
        Check the mean value of the FormFactor of the nucleus to determine
        if whole plate or partial plate contamination is present.
        Threshold is set based on findings from multiple experiments
        (NF1 & CFReT, see citation file).

        Normal raw values of this feature range from 0 to 1. In a clean plate,
        it is noted that the normal values range from 0.8 to 1 (heavily skewed).
        We set the threshold to 0.78 to detect whole plate contamination.

        Returns:
            boolean: True if whole plate contamination is detected, False otherwise
            indicates partial contamination.
        """
        # check if the mean value is below the threshold
        # True = whole plate, False = partial plate
        return (
            self.dataframe[self.formfactor_feature].mean()
            <= self.formfactor_mean_threshold
        )

    def check_feature_means(self) -> None:
        """
        Step 2: If skewness and/or variability detected, determine if the whole plate or
        partial plate is impacted using the mean values of the features.

        Main interpretations:
        - If whole plate contamination detected in both texture and formfactor
            → major warning.
        - If whole plate contamination detected in texture only
            → check nucleus channel images for contamination.
        - If whole plate contamination detected in formfactor only
            → check segmentation parameters and/or images for contamination.
        - If partial plate contamination detected in texture only
            → proceed to step 3.
        - If partial plate contamination detected in formfactor only
            → recommend performing 'find_outliers' with FormFactor.
        """
        # run sanity check to ensure step one was ran prior to step two
        if not hasattr(self, "is_skewed") or not hasattr(self, "is_variable"):
            raise RuntimeError(
                "You must run the skew and variable test before performing this test."
            )

        # catch case where steps are used individually and no contamination is present
        if not self.is_skewed and not self.is_variable:
            print("Data is neither skewed nor variable based on step one.")
            print("No indication of contamination. Plate appears clean 🫧.")
            self.partial_contamination_texture_detected = False
            return

        print("Running step 2...")

        # Instantiate variables to check for whole plate contamination
        self.whole_plate_contamination_texture = (
            self._calculate_texture_mean() if self.is_skewed else False
        )
        self.whole_plate_contamination_formfactor = (
            self._calculate_formfactor_mean() if self.is_variable else False
        )

        # Set lookup dictionary for all interpretations based on the boolean values
        interpretation_lookup = {
            (
                True,
                True,
                True,
                True,
            ): (
                "MAJOR WARNING! 💥\n"
                "Contamination across entire plate detected in both texture and nuclei "
                "shape features. Strongly suggest inspecting nucleus channel images "
                "for contamination."
            ),
            (
                True,
                False,
                True,
                False,
            ): (
                "Whole plate texture contamination detected. "
                "Please check your nucleus channel images."
            ),
            (
                False,
                True,
                False,
                True,
            ): (
                "Whole plate shape contamination detected. "
                "Check segmentation parameters and/or images."
            ),
            (
                True,
                True,
                True,
                False,
            ): (
                "Whole plate texture + partial plate shape contamination. "
                "Inspect nucleus channel and segmentation in a subset of FOVs."
            ),
            (
                True,
                True,
                False,
                False,
            ): (
                "Partial contamination detected in both texture and shape. "
                "Proceed to step 3 and inspect FormFactor outliers separately."
            ),
            (
                True,
                False,
                False,
                False,
            ): "Partial plate contamination in texture only. Proceed to step 3.",
            (
                False,
                True,
                False,
                False,
            ): (
                "Partial shape deviation detected. "
                "Recommend running 'find_outliers' with FormFactor."
            ),
        }

        # Set up key for lookup mapping to be applied
        key = (
            self.is_skewed,  # checks texture contamination
            self.is_variable,  # checks shape contamination
            self.whole_plate_contamination_texture,
            self.whole_plate_contamination_formfactor,
        )

        # Get the interpretation from the lookup dictionary
        interpretation = interpretation_lookup.get(
            key,
            (
                "Combination not explicitly handled. "
                "Consider inspecting both channels manually."
            ),
        )

        # Format interpretation as table
        print("Summary:\n")
        print(
            tabulate(
                [
                    ["Texture skewed?", bool_to_emoji(self.is_skewed)],
                    ["Nucleus shape variable?", bool_to_emoji(self.is_variable)],
                    [
                        "Whole plate contaminated due to abnormal texture?",
                        bool_to_emoji(self.whole_plate_contamination_texture),
                    ],
                    [
                        "Whole plate contaminated due to abnormal nucleus shape?",
                        bool_to_emoji(self.whole_plate_contamination_formfactor),
                    ],
                ],
                headers=["Check", "Result"],
                tablefmt="simple",
            )
        )
        print("Interpretation:\n" + interpretation)

        # Set flag used in step 3
        self.partial_contamination_texture_detected = (
            self.is_skewed and not self.whole_plate_contamination_texture
        )

    def _find_texture_outliers(self) -> pd.DataFrame:
        """
        Use coSMicQC find_outliers function to identify contaminated single-cells based
        on texture in the cytoplasm around the nucleus in partially contaminated plates.

        Returns:
            pd.DataFrame: DataFrame containing the outliers detected to use to plotting
            and calculating proportions.
        """
        # necessary metadata columns for processing
        metadata_columns = [
            "Image_Metadata_Plate",
            "Image_Metadata_Well",
            "Image_Metadata_Site",
            "Metadata_Nuclei_Location_Center_X",
            "Metadata_Nuclei_Location_Center_Y",
        ]

        # detect outliers based on original texture feature
        texture_feature_thresholds = {
            # outlier threshold for only cytoplasm texture in nuclei channel
            self.cyto_feature: self.outlier_std_threshold,
        }

        texture_nuclei_outliers = find_outliers(
            df=self.dataframe,
            metadata_columns=metadata_columns,
            feature_thresholds=texture_feature_thresholds,
        )

        # for added support, also check for outliers using granularity of the cytoplasm
        # in the nuclei channel
        granularity_feature_thresholds = {
            # outlier threshold for only cytoplasm granularity in nuclei channel
            f"Cytoplasm_Granularity_2_{self.nucleus_channel_naming}": self.outlier_std_threshold,  # noqa: E501
        }

        granularity_nuclei_outliers = find_outliers(
            df=self.dataframe,
            metadata_columns=metadata_columns,
            feature_thresholds=granularity_feature_thresholds,
        )

        # combine the two outlier dataframes
        combined_outliers = pd.concat(
            [texture_nuclei_outliers, granularity_nuclei_outliers]
        ).drop_duplicates(
            subset=[
                "Image_Metadata_Well",
                "Image_Metadata_Site",
                "Metadata_Nuclei_Location_Center_X",
                "Metadata_Nuclei_Location_Center_Y",
            ],
            keep="first",
        )

        # print total number of outliers detected
        print(f"Total number of outliers detected: {len(combined_outliers)}")

        # return the outliers dataframe
        return combined_outliers

    def _get_outlier_proportion_per_well(self) -> pd.DataFrame:
        """
        Calculate the proportion of outliers per well to be used for plotting.

        Returns:
            pd.DataFrame: Proportion per well dataframe
        """
        # Get the outliers detected
        combined_outliers = self._find_texture_outliers()

        # Calculate the proportion of outliers per well
        outlier_proportion_per_well = (
            combined_outliers.groupby("Image_Metadata_Well")
            .size()
            .div(self.dataframe.groupby("Image_Metadata_Well").size(), fill_value=0)
            .fillna(0)
            .reset_index()
        )
        outlier_proportion_per_well.columns = ["Well", "Proportion"]
        outlier_proportion_per_well["Proportion"] *= 100

        # Sort the wells in order of row and column (e.g., A01, A02, B01, B02, ...)
        outlier_proportion_per_well["Well"] = pd.Categorical(
            outlier_proportion_per_well["Well"],
            sorted(
                outlier_proportion_per_well["Well"], key=lambda x: (x[0], int(x[1:]))
            ),
        )
        outlier_proportion_per_well = outlier_proportion_per_well.sort_values("Well")

        # Calculate the cell counts per well
        cell_count_per_well = (
            self.dataframe.groupby("Image_Metadata_Well")
            .size()
            .reset_index(name="CellCount")
        )

        # Merge cell count with proportions
        outlier_proportion_per_well = outlier_proportion_per_well.merge(
            cell_count_per_well, left_on="Well", right_on="Image_Metadata_Well"
        )

        return outlier_proportion_per_well

    def _plot_outlier_proportions(self, df: pd.DataFrame) -> None:
        """
        Plot the proportion of outliers per well with a color gradient
        based on the total cell count per well.

        Args:
            df (pd.DataFrame): Proportion per well DataFrame
        """
        # Normalize the cell count for color mapping
        norm = plt.Normalize(df["CellCount"].min(), df["CellCount"].max())
        colors = plt.cm.viridis(norm(df["CellCount"]))

        # Create bar plot
        plt.figure(figsize=(20, 6))
        ax = sns.barplot(
            data=df,
            x="Well",
            y="Proportion",
            hue="Well",
            palette=list(colors),
            legend=False,
        )
        plt.xlabel("Well")
        plt.ylabel("Proportion")
        plt.xticks(rotation=45)
        plt.ylim(0, 100)

        sm = plt.cm.ScalarMappable(cmap="viridis", norm=norm)
        sm.set_array([])
        plt.colorbar(sm, ax=ax, label="Total Cell Count")
        plt.show()

    def check_partial_contamination(self) -> None:
        """
        Step 3: If partial contamination detected, find outliers and plot the proportion
        of outliers per well.
        """
        # run sanity check to ensure step two was ran prior to step three
        if not hasattr(self, "partial_contamination_texture_detected"):
            raise RuntimeError(
                "You must run the feature mean test before performing this test."
            )

        print("Running step 3...")

        # check if partial contamination was detected
        if self.partial_contamination_texture_detected:
            print("Finding outlier cells with anomalous texture around the nucleus...")
            outliers = self._get_outlier_proportion_per_well()
            self._plot_outlier_proportions(outliers)

            # Calculate the 75th percentile (top 25%) of the outlier proportions
            top_25_percent_threshold = outliers["Proportion"].quantile(0.75)

            # Filter wells with proportions in the top 25%
            top_outlier_wells = outliers[
                outliers["Proportion"] >= top_25_percent_threshold
            ]
            # Print the wells in the top 25% horizontally
            print(f"\nNumber of wells in the top 25%: {len(top_outlier_wells)}")
            print("Wells in the top 25% of highest outlier proportions:")
            print(", ".join(top_outlier_wells["Well"]))
        else:
            print("No partial contamination detected; no outliers to find.")

    def run(self) -> None:
        """
        Run all steps of the contamination detection process with conditional logic.
        """
        # Run the first step in the process to check for skewness and variability
        self.check_skew_and_variable()

        # Exit early if no skewness or variability detected
        if not self.is_skewed and not self.is_variable:
            print(
                "No skewness or variability detected. Exiting contamination detector."
            )
            return

        # Run the second step if skewness or variability was detected
        # Determines if the whole plate or partial plate is impacted
        self.check_feature_means()

        # Only continue to step 3 if partial contamination texture was detected
        if (
            hasattr(self, "partial_contamination_texture_detected")
            and self.partial_contamination_texture_detected
        ):
            self.check_partial_contamination()



"""
Initialization for cosmicqc package
"""

from .analyze import find_outliers, identify_outliers, label_outliers
from .detection import ContaminationDetector

# note: version placeholder is updated during build
# by poetry-dynamic-versioning.
__version__ = "1.0.1"
